/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package com.ebmwebsourcing.jbi.adaptor.impl.mock;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchange.Role;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.mockito.Mockito;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.jbi.adaptor.impl.exchange.ContextualMessageExchangeFactoryImpl;

/**
 * @author Adrien Ruffie - EBM WebSourcing
 */
public class MockDeliveryChannel implements DeliveryChannel {

    private final Logger logger;

    private final LinkedBlockingQueue<MessageExchange> incomingMessagesQueue;

    private final LinkedBlockingQueue<MessageExchange> responsesToClientQueue;

    private final LinkedBlockingQueue<MessageExchange> requestsToServerQueue;

    private String synchronousExchangeId;

    private final Semaphore synchronousExchangeSemaphore = new Semaphore(0);

    public MockDeliveryChannel(final Logger logger) {
        this.logger = logger;
        this.incomingMessagesQueue = new LinkedBlockingQueue<MessageExchange>();
        this.responsesToClientQueue = new LinkedBlockingQueue<MessageExchange>();
        this.requestsToServerQueue = new LinkedBlockingQueue<MessageExchange>();
    }

    public void enqueueIncomingMessage(final MessageExchange msgExchange) {
        try {
            this.incomingMessagesQueue.put(msgExchange);
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    public void enqueueRequestToServer(final MessageExchange msgExchange) {
        try {
            this.requestsToServerQueue.put(msgExchange);
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    public void enqueueResponseToClient(final MessageExchange msgExchange) {
        try {
            this.responsesToClientQueue.put(msgExchange);
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    public MessageExchange dequeueIncomingMessage() {
        try {
            return this.incomingMessagesQueue.take();
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    public MessageExchange dequeueRequestToServer() {
        try {
            return this.requestsToServerQueue.take();
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    public MessageExchange dequeueResponseToClient() {
        try {
            return this.responsesToClientQueue.take();
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    @Override
    public MessageExchange accept() throws MessagingException {
        try {
            MessageExchange exchange = this.incomingMessagesQueue.take();
            if (exchange.getExchangeId().equals(synchronousExchangeId)) {
                synchronousExchangeSemaphore.release();
                return accept();
            } else {
                return exchange;
            }
        } catch (final InterruptedException e) {
            throw new UncheckedException(e);
        }
    }

    @Override
    public MessageExchange accept(final long timeoutMS) throws MessagingException,
            IllegalArgumentException {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public void close() throws MessagingException {
        // TODO Auto-generated method stub
    }

    @Override
    public MessageExchangeFactory createExchangeFactory() {
        final ContextualMessageExchangeFactoryImpl exchangeFactoryImpl = ContextualMessageExchangeFactoryImpl
                .createContextualMessageExchangeFactory(Mockito.mock(ServiceEndpoint.class),
                        this.logger);
        return exchangeFactoryImpl;
    }

    @Override
    public MessageExchangeFactory createExchangeFactory(final QName interfaceName) {
        final ContextualMessageExchangeFactoryImpl exchangeFactoryImpl = ContextualMessageExchangeFactoryImpl
                .createContextualMessageExchangeFactoryForInterface(
                        Mockito.mock(ServiceEndpoint.class), this.logger, interfaceName);
        return exchangeFactoryImpl;
    }

    @Override
    public MessageExchangeFactory createExchangeFactory(final ServiceEndpoint endpoint) {
        final ContextualMessageExchangeFactoryImpl exchangeFactoryImpl = ContextualMessageExchangeFactoryImpl
                .createContextualMessageExchangeFactoryForEndpoint(
                        Mockito.mock(ServiceEndpoint.class), this.logger, endpoint);
        return exchangeFactoryImpl;
    }

    @Override
    public MessageExchangeFactory createExchangeFactoryForService(final QName serviceName) {
        final ContextualMessageExchangeFactoryImpl exchangeFactoryImpl = ContextualMessageExchangeFactoryImpl
                .createContextualMessageExchangeFactoryForService(
                        Mockito.mock(ServiceEndpoint.class), this.logger, serviceName);
        return exchangeFactoryImpl;
    }

    @Override
    public void send(final MessageExchange messageExchange) throws MessagingException {
        if (messageExchange.getEndpoint() == null) {
            resolveEndpoint(messageExchange);
        }

        try {
            if (messageExchange.getRole().equals(Role.PROVIDER)) {
                this.responsesToClientQueue.put(messageExchange);
            } else {
                this.requestsToServerQueue.put(messageExchange);
            }
        } catch (final InterruptedException e) {
            new UncheckedException(e);
        }
    }

    private void resolveEndpoint(MessageExchange messageExchange) throws MessagingException {
        ServiceEndpoint endpoint = MockComponentContext.resolveServiceEndpoint(messageExchange
                .getService());
        if (endpoint != null) {
            messageExchange.setEndpoint(endpoint);
        } else {
            throw new MessagingException("Endpoint not avalaible");
        }
    }

    @Override
    public boolean sendSync(final MessageExchange msgExchange) throws MessagingException {
        return this.sendSync(msgExchange, 0);
    }

    @Override
    public boolean sendSync(final MessageExchange msgExchange, final long timeoutMS)
            throws MessagingException {
        boolean returnValue = false;
        synchronousExchangeId = msgExchange.getExchangeId();
        this.send(msgExchange);
        try {
            if (timeoutMS > 0) {
                returnValue = synchronousExchangeSemaphore.tryAcquire(timeoutMS,
                        TimeUnit.MILLISECONDS);

            } else {
                synchronousExchangeSemaphore.acquire();
                returnValue = true;
            }
        } catch (InterruptedException e) {
            new UncheckedException(e);
        }
        return returnValue;
    }

}
